/**
 * Breakout plays the classic game of making a bouncing
 * ball bounce off a paddle and bricks until the ball 
 * falls off the bottom of the scene. In this lab you
 * will modify the implementation and REPORT THE CHANGES
 * IN THIS HEADER COMMENT.
 * 
 * The changes you should make to this file are Tasks 1
 * through 4, listed in reverse order.  Start with Task 1.
 */
import acm.graphics.*;
import acm.program.*;
import java.awt.event.*;
import java.awt.*;

// RandomGenerator is a class in acm.util. Read about
// it in the acm student package documentation (there
// is a link to that documentation on the course wiki.
// In this file, it sets the initial x coordinate
// of the ball to a random integer (on the first line
// of the createBall method.) You can use the same 
// RandomGenerator object to generate random integers,
// doubles and even Colors.
import acm.util.RandomGenerator;

public class Breakout extends GraphicsProgram {
    
    // private instance variables:
    // random number generator rgen created from acm.jar
    // class RandomGenerator
    private RandomGenerator rgen = RandomGenerator.getInstance();
    private GOval ball;
    private boolean isInPlay;
    private GRect paddle;
    private double dx = 2.0;
    private double dy = 2.0;
    // TASK 4, part 1:  ADD A GRECT FOR THE BRICK
    
    
    // public named constants.  READ THESE
    public static final int BALL_SIZE = 20;
    public static final int PADDLE_HEIGHT = 10;
    public static final int PADDLE_WIDTH = 100;
    public static final double PAUSE_TIME = 10.0;  
    public static final int APPLICATION_HEIGHT = 600;
    public static final int APPLICATION_WIDTH = 400;
    
    /**
     * run: 
     *
     * Invoked by the start method in main, 
     * causes execution to continue from here.
     */
    public void run () {
        setUpGame();
        runGame();
    }
    
    /**
     * setUpGame: 
     * 
     * TASK 4, part 2:
     * 
     * Add a "brick" near the top of the canvas that will 
     * disappear when the ball contacts it.  For now, make 
     * the brick almost as wide as the canvas so the ball 
     * can't miss it. The ball should bounce off this brick 
     * in the same way it bounces off the paddle, only in 
     * this case the brick should disappear. This will 
     * require a change to the checkForObjectBounces method 
     * and to the setUpGame method. The brick and the paddle 
     * should appear when the game starts.
     *
     * Called by the run method to set up the initial scene.
     * Creates the unmoving paddle and adds MouseListeners
     */
    private void setUpGame() {  
        paddle = new GRect(getWidth()/2 - PADDLE_WIDTH/2,
                           getHeight() - PADDLE_HEIGHT/2,
                           PADDLE_WIDTH, PADDLE_HEIGHT);
        paddle.setFilled(true);
        add(paddle);
        addMouseListeners();
    }
    
    /**
     * createBall: 
     * 
     * TASK 3: Change starting position of ball so that it
     * starts midway between the top and bottom of the scene.
     * This only requires a change to the y-coordinate of the
     * ball.
     *
     * Called by the runGame method to create a new GOval on 
     * the scene to be the bouncing ball. Returns the ball that 
     * will bounce around the window.
     * 
     * @return ball 
     */
    private GOval createBall() {
        int x = rgen.nextInt(0,getWidth()-BALL_SIZE);
        GOval tempBall =  new GOval(x,0,BALL_SIZE,BALL_SIZE);
        tempBall.setFilled(true);
        return tempBall;
    } 
    
    /**
     * runGame:  (NO CHANGE NEEDED)
     *
     * Called by the run method to display and remove 
     * the initial instructions, and add the ball to 
     * the canvas. Calls the method playBall to start 
     * moving the ball.
     */
    private void runGame() {
        GLabel message = 
            new GLabel("Click to start!",PADDLE_WIDTH,PADDLE_WIDTH);
        message.setFont("SansSerif-36");
        add(message);
        waitForClick();
        remove(message);
        ball = createBall();
        ball.setFillColor(Color.RED);
        add(ball);     
        playBall();
    }
    
    /**
     * playBall:  (NO CHANGE NEEDED)
     * 
     * isInPlay is an initially true boolean indicating 
     * the ball has dropped off the bottom of the scene, 
     * when isInPlay is set to false
     * 
     * Moves the ball dx pixels in the x direction
     * and dy pixels in the y direction, pauses so 
     * the ball can be redrawn and then calls
     * methods to check for wall and paddle collisions.
     */
    private void playBall() {
        while (isInPlay) {
            ball.move(dx,dy);
            pause(PAUSE_TIME);
            checkForWallBounces();
            checkForObjectBounces();
        }
    }
    
    /**
     * checkForWallBounces: (Change this method)
     * 
     * If ball is in collision with one of the walls 
     * change the direction of ball movement by either 
     * reversing dx or dy
     * 
     * TASK 2: Modify this method so ball does not bounce 
     * off bottom of scene
     *
     * TASK 5: When the ball falls off the bottom of the
     * scene, display the message "GAME OVER", followed
     * by the user's score: the number of times the ball
     * bounced off the paddle before dropping off the
     * bottom of the scene. To see how to add a GLabel,
     * look at the runGame method, above.
     * 
     * The GAME OVER GLabel should be above the Glabel
     * "Your score is XX!", both roughtly centered 
     * horizontally on the scene.
     */
    private void checkForWallBounces() {
        double x = ball.getX();
        double y = ball.getY();
        if (x < 0 || x + BALL_SIZE > getWidth()) dx = -dx; 
        if (y < 0 || y + BALL_SIZE > getHeight()) dy = -dy;
    }
    
    /**
     * checkForObjectBounces: (Change this method)
     *
     * If ball is in collision with the paddle, this
     * method changes the direction of ball movement
     * by negating the dy variable.
     * 
     * TASK 4, part 3:
     * Change this method so that, if the ball hits
     * the brick, the brick disappears and the ball 
     * negates dy variable.
     */
    private void checkForObjectBounces() {
        GObject obj = getCollidingObject();
        if (obj == null) return;
        if (obj == paddle) {
            dy = -dy;       
        }
    }
    
    /**
     * getCollidingObject: (NO CHANGE NEEDED)
     * 
     * Returns the GObject that the ball is currently
     * touching by checking all 4 corners of the ball
     * for contact.
     */
    private GObject getCollidingObject() {
        double x = ball.getX();
        double y = ball.getY();
        double d = BALL_SIZE;
        GObject obj = getElementAt(x,y);  // upper left corner
        if (obj == null) obj = getElementAt(x+d,y); // upper right
        if (obj == null) obj = getElementAt(x,y+d); // lower left
        if (obj == null) obj = getElementAt(x+d,y+d);// lower right
        return obj;
    }
    
    /** 
     * mouseClicked:  (NO CHANGE NEEDED)
     *
     * Side effect is setting the boolean to start the
     * ball moving to true
     */
    public void mouseClicked(MouseEvent e) {
        isInPlay = true;
    }
    
    /**
     * mouseMoved: (TASK 1: Write the mouseMoved method)
     * 
     * Add this method as step one of the program 
     * specifications. In this method, the paddle is 
     * setup to follow the mouse moving on the scene.
     */
    
    
    
    /**
     * main:   (NO CHANGE NEEDED)
     *
     * Calls start method of Program class that in turn causes
     * a hidden call to the run method in this class
     */
    public static void main(String[] args) {
        new Breakout().start();
    }
}